#!/usr/bin/perl
#
# $Revision: 1.47 $
# Copyright 2009 Teleflora
# 
# Setup a "parallel" SSHd service which only permits 2-factor authentication.
# PA-DSS 11.2
# PCI-DSS 8.2, 8.3
#

use strict;
use warnings;
use Digest::MD5;
use Getopt::Long;
use POSIX;
use English;

my $OS = "";
my $VERSION = 0;
my $HELP = 0;
my $INSTALL = 0;
my $START = 0;
my $STOP = 0;
my $STATUS = 0;
my $MAKEKEYS = 0;
my $ENABLE = "";
my $DISABLE = "";
my $CONNECT = "";
my $ALLOW_FROM = "";
my $ACCEPT = 0;

GetOptions(
	"version" => \$VERSION,
	"help" => \$HELP,
	"install" => \$INSTALL,
	"start" => \$START,
	"stop" => \$STOP,
	"status" => \$STATUS,
	"connect=s" => \$CONNECT,
	"enable-tfremote=s" => \$ENABLE,
	"disable-tfremote=s" => \$DISABLE,
	"allow-from=s" => \$ALLOW_FROM,
	"accept" => \$ACCEPT,
);



# --help
if($HELP != 0) {
	usage();
	exit(0);
}

# --version
if($VERSION != 0) {
	print("OSTools Version: 1.12.5\n");
	print('$Revision: 1.47 $' . "\n");
	exit(0);
}

# --status
if($STATUS != 0) {
	exit(tfremote_status());
}


# --connect some.host.com
# --connect=some.host.com
if("$CONNECT" ne "") {
	exit(connect_to_host($CONNECT));
}


# --enable-tfremote username
# --disable-tfremote username
if("$ENABLE" ne "") {
	my @array = ();
	if("$ALLOW_FROM" ne "") {

		# --allow-from=192.168.*,10.*,127.0.0.1 -> @array
		chomp($ALLOW_FROM);
		@array = split(/,/, $ALLOW_FROM);
	}
	exit(enable_tfremote($ENABLE, @array));
}
if("$DISABLE" ne "") {
	exit(disable_tfremote($ENABLE));
}



# --accept
if($ACCEPT != 0) {
	exit(accept_inbound_connection());
}


############
# Root is needed to execute commands below.
############
if($< != 0) {
	usage();
	print("Error: Must run this script as 'sudo' or root.\n");
	exit(1);
}


# --start
if($START != 0) {
	exit(start_tfremote());
}

# --stop
if($STOP != 0) {
	exit(stop_tfremote());
}

# --install
if($INSTALL != 0) {
	exit(install_tfremote());
}




usage();

exit(0);
#####################################################################
#####################################################################
#####################################################################


sub usage
{
	print("Usage:\n");
	print("$0 --help\n");
	print("$0 --version\n");
	print("$0 --start\n");
	print("$0 --stop\n");
	print("$0 --status\n");
	print("$0 --install\n");
	print("$0 --accept\n");
	print("$0 --connect some.host.com\n");
	print("$0 --connect user\@sftp.teleflora.com\n");
	print("$0 [--allow-from=host.ip.address,*.ip.address,host.ip.*,...] --enable-tfremote username\n");
	print("$0 --disable-tfremote username\n");
	print("\n");
	print("TFRemote is a system service which only allows incoming connections authenticated\n");
	print("via ssh public key encryption. As the service is actually a second instance of OpenSSH,\n");
	print("all configuration and usage rules which apply to SSH, apply here as well.\n");
	print("This script is capable of setting up the 'tfremote' service\n");
	print("on a redhat machine, as well as starting and stopping the service.\n");
	print("\n");
	return(0);
}




#
# RHEL 5, and later versions of RHEL 4 support "appending" group memberships. However,
# Early (initial release) versions of RHEL4 do not support "-a". This is a sort of 'hand rolled'
# "append" flag.
# Yes, this is pretty inefficient, but then it's not used very often either.
#
sub append_usergroup
{
	my $newgroup = $_[0];
	my $username = $_[1];
	my $line = "";

	if($newgroup eq "") {
		return(-1);
	}
	if($username eq "") {
		return(-2);
	}

	# Get a list of current groups.
	open(PIPE, "groups $username |");
	$line = <PIPE>;
	close(PIPE);

	# "root : root foo bar fee" -> "root foo bar fee"
	$line =~ s/^([[:print:]]+)(:\s+)([[:print:]]+)/$3/g;

	# We are already a member of this group.
	if($line =~ /(\s+)($newgroup)/) {
		return;
	}

	# "root foo bar fee" -> "root foo bar fee baz"
	$line .= " $newgroup";

	# "root foo bar fee baz" -> "root,foo,bar,fee,baz"
	$line =~ s/(\s+)/,/g;


	# Actually set groups here.
	system("/usr/sbin/usermod -G \"$line\" $username");


	return;
}


#
# Get useful information such as groups and home directory
# about a given user.
sub get_userinfo
{
	my $username = $_[0];
	my $groupname = "";
	my $users = "";
	my @ent = ();
	my @groups = ();
	my %hash;
	my $found = 0;

	if($username eq "") {
		return %hash;
	}

	# Get user related info.
	setpwent();
	@ent = getpwent();
	while(@ent) {
		if( "$ent[0]" eq "$username") {
			$found = 1;
			$hash{"username"} = $username;
			$hash{"homedir"} = "$ent[7]";
			$hash{"shell"} = "$ent[8]";
			last;
		}
		@ent = getpwent();
	}
	endpwent();
 

	# We did  not find this user.
	if($found == 0) {
		return %hash;
	}


	# Get user's associated groups.
	# ($name,$passwd,$gid,$members) = getgr*
	$hash{"groups"} = ();
	@ent = getgrent();
	while(@ent) {
		$groupname = $ent[0];
		$users = $ent[3];
		if($users =~ /([[:space:]]*)($username)([[:space:]]{1}|$)/) {
			push @groups, $groupname;
		}
		@ent = getgrent();
	}
	endgrent();
	$hash{'groups'} = [ @groups ];
 

	return(%hash);
}


#
# Taken from rtiuser.pl rev 1.65.
#
sub remove_usergroup
{
	my $delgroup= $_[0];
	my $username = $_[1];
	my @array = ();
	my $line = "";



	# Get a list of current groups.
	open(PIPE, "groups $username |");
	$line = <PIPE>;
	close(PIPE);


	# "root : root foo bar fee" -> "root foo bar fee"
	$line =~ s/^([[:print:]]+)(:)//g;


	# "root foo bar fee" -> "root bar fee"
	$line =~ s/(\s+)($delgroup)//g;

	# "    root bar fee    " -> "root bar fee"
	$line =~ s/^(\s+)//g;
	$line =~ s/(\s+)$//g;


	# "root bar fee" -> "root,bar,fee"
	$line =~ s/\s+/,/g;

	system("/usr/sbin/usermod -G \"$line\" $username");
}

sub start_tfremote
{
	my $prog = "";

	# Make sure the service starts with each reboot.
	foreach $prog ("/sbin/chkconfig", "/usr/sbin/chkconfig") {
		next if(! -f "$prog");
		system("$prog --add tfremote");
		system("$prog --level 35 tfremote on");
		last;
	}

	exec("/sbin/service tfremote start");
}


sub stop_tfremote
{
	my $prog = "";

	# Make sure the service does not start on reboot.
	foreach $prog ("/sbin/chkconfig", "/usr/sbin/chkconfig") {
		next if(! -f "$prog");
		system("$prog --level 35 tfremote off");
		last;
	}

	exec("/sbin/service tfremote stop");
}


sub tfremote_status
{
	exec("/sbin/service tfremote status");
}




#
# Create the system service.
# This is basically a clone of "sshd" which runs concurrent to sshd,
# uses the same config files, but is just renamed.  The reason for
# a separate, concurrent process, is to enforce on a "system" level,
# the use of two-factor authentication.
#
sub install_tfremote
{
	my $prog = "";
	my $file = "";
	my $string = "";


	# Verify our "source files" exist.
	foreach $file ("/etc/init.d/sshd", "/etc/ssh/sshd_config", "/usr/sbin/sshd") {
		unless (-f "$file") {
			print("Error: \"$file\". $!\n");
			return(1);
		}
	}

	foreach $file ("/etc/sysconfig", "/etc/ssh") {
		unless (-d "$file") {
			print("Error: \"$file\" $!\n");
			return(2);
		}
	}

	#
	# The presense of the 'tfremote' group is required.  Try to add it, but
	# if it's already there, it's ok.
	#
	system("sudo /usr/sbin/groupadd tfremote > /dev/null 2>&1 /dev/null");

	# /etc/init.d
	open(OLD, "< /etc/init.d/sshd");
	open(NEW, "> /etc/init.d/tfremote");
	while(<OLD>) {
		$string = $_;
		$string =~ s/ssh_host_key/tfremote_host_key/g;
		$string =~ s/ssh_host_rsa_key/tfremote_host_rsa_key/g;
		$string =~ s/ssh_host_dsa_key/tfremote_host_dsa_key/g;

		$string =~ s/ssh_random_seed/tfremote_random_seed/g;
		$string =~ s/sshd/tfremote/g;
		$string =~ s/prog=\"([[:print:]]+)\"/prog=\"tfremote\"/g;
		$string =~ s/description: OpenSSH server daemon/description:Teleflora POS Remote Access/g;

		print(NEW $string);
	}
	close(OLD);
	close(NEW);
	system("chmod --reference=/etc/init.d/sshd /etc/init.d/tfremote");
	system("chown --reference=/etc/init.d/sshd /etc/init.d/tfremote");


	system("mkdir -p /var/empty/tfremote");
	system("chmod --reference=/var/empty/sshd /var/empty/tfremote");
	system("chown --reference=/var/empty/sshd /var/empty/tfremote");


	# Make the service start at bootup.
	# Must use a hard link here (not symlink), as, redhat init.d/functions "stop()" function will
	# wind up killing all "sshd" processes if we use a symlink.
	# Redhat updates should work fine in the event that we use hard links.
	unlink("/usr/sbin/tfremote");
	#
	# FIXME: using hard links is an issue for updates.
	#
	system("ln /usr/sbin/sshd /usr/sbin/tfremote");


	# tfremote_config
	open(OLD, "< /etc/ssh/sshd_config");
	open(NEW, ">  /etc/ssh/tfremote_config");
	while(<OLD>)
	{
		next if(/Port /i);
		next if(/ListenAddress /i);
		next if(/Protocol /i);
		next if(/HostKey /i);
		next if(/AcceptEnv /i);
		next if(/SyslogFacility /i);
		next if(/LogLevel /i);
		next if(/LoginGraceTime /i);
		next if(/MaxAuthTries /i);
		next if(/MaxStartups /i);
		next if(/PidFile /i);
		next if(/PermitRootLogin /i);
		next if(/RSAAuthentication /i);
		next if(/PubkeyAuthentication /i);
		next if(/RhostsRSAAutthenitcation /i);
		next if(/HostbasedAuthentication /i);
		next if(/IgnoreUserKnownHosts /i);
		next if(/IgnoreRhosts /i);
		next if(/PermitEmptyPasswords /i);
		next if(/PasswordAuthentication /i);
		next if(/KerberosAuthentication /i);
		next if(/UsePrivilegeSeparation /i);
		next if(/GSSAPIAuthentication /i);
		next if(/UsePAM /i);
		next if(/Banner /i);
		next if(/ClientAliveInterval /i);
		next if(/ClientAliveCountMax /i);

		print(NEW);
	}
	close(OLD);


	print(NEW "\n");
	print(NEW "\n");
	print(NEW "#\n"); 
	print(NEW "# PA-DSS Compliant Settings\n");
	print(NEW "# Setup by tfremote " . '$Revision: 1.47 $' . "\n");
	print(NEW "#\n"); 
	print(NEW "Port 15022\n");
	print(NEW "ListenAddress 0.0.0.0\n");
	print(NEW "Protocol 2\n");
	print(NEW "HostKey /etc/ssh/tfremote_host_rsa_key\n");
	print(NEW "HostKey /etc/ssh/tfremote_host_dsa_key\n");
	print(NEW "AcceptEnv BBTERM\n");
	print(NEW "SyslogFacility AUTH\n");
	print(NEW "LogLevel VERBOSE\n");
	print(NEW "LoginGraceTime 30\n");
	print(NEW "MaxAuthTries 1\n");
	print(NEW "MaxStartups 3:30:10\n");
	print(NEW "PidFile /var/run/tfremote.pid\n");
	print(NEW "PermitRootLogin no\n");
	print(NEW "RSAAuthentication no\n");
	print(NEW "PubkeyAuthentication yes\n");
	print(NEW "RhostsRSAAuthentication no\n");
	print(NEW "HostBasedAuthentication no\n");
	print(NEW "IgnoreUserKnownHosts yes\n");
	print(NEW "IgnoreRhosts yes\n");
	print(NEW "PermitEmptyPasswords no\n");
	print(NEW "PasswordAuthentication no\n");
	print(NEW "KerberosAuthentication no\n");
	print(NEW "UsePrivilegeSeparation yes\n");
	print(NEW "GSSAPIAuthentication no\n");
	print(NEW "UsePAM no\n");
	print(NEW "DenyGroups *\n");
	print(NEW "DenyUsers rti daisy\n"); # PCI-DSS 8.1
	print(NEW "AllowGroups tfremote\n");
	print(NEW "AllowTcpForwarding yes\n");
	print(NEW "X11Forwarding no\n");
	print(NEW "Banner /etc/motd\n");
	print(NEW "ClientAliveInterval 300\n");
	print(NEW "ClientAliveCountMax 3\n");
	print(NEW "\n");
	print(NEW "\n");
	close(NEW);

	unless (-f "/etc/ssh/tfremote_config") {
		print("Error: /etc/ssh/tfremote_config  $!\n");
		return(5);
	}


	# Here is where we tell 'sshd' to read our PA-DSS compliant settings.
	open(NEW, "> /etc/sysconfig/tfremote");
	print(NEW "OPTIONS=\"-f /etc/ssh/tfremote_config\"\n");
	close(NEW);

	unless (-f "/etc/sysconfig/tfremote") {
		print("Error: /etc/sysconfig/tfremote $!\n");
		return(6);
	}
}







#
# Connect to a tfremote server.
# Doubles as a "sftp client" to help us connect to sftp.teleflora.com
# This script is ued mainly on our customer service servers to connect to customer machines; setting up appropriate security
# parameters, as well as port forwarding rules.
#
sub connect_to_host
{
	my $hostname = $_[0];
	my %userinfo = ();
	my $localuser = getpwuid($>);
	my $command = "";
	my $portnum = int(10000 + int($UID));
	my $remoteuser = "";
	my $id_dsa = "";
	my $thisfile = "";
	my $fingerprint = "";
	my @array = ();


	# Who is the current user?
	if ($localuser eq "") {
		system("/usr/bin/logger Could not determine name of effective user.");
		print("Error: Could not determine name of effective user.\n");
		return (1);
	}

	# We want to sftp out.
	# In some cases, we have had issues where the tfsupport user account has an 'id_dsa' or 'id_rsa' in their .ssh directory
	# With default "sftp", these private keys are presented to sftp.teleflora.com, who denies our login and fails to allow a 
	# failover to password.
	# Thus, this scriptlet allows us to skip the pubkey authentication and only use password auth.
	if("$hostname" =~ /sftp\.teleflora\.com/i) {
		if($hostname =~ /(\S+)(\@)(sftp)/) {
			system("/usr/bin/sftp -o PubKeyAuthentication=no $hostname");
		} else {
			if(-d "/d/daisy") {
				system("/usr/bin/sftp -o PubKeyAuthentication=no daisydev\@sftp.teleflora.com");
			} else {
				system("/usr/bin/sftp -o PubKeyAuthentication=no rtidev\@sftp.teleflora.com");
			}
		}
		return($?);
	}




	# Where is our home directory?
	%userinfo = get_userinfo($localuser);
	foreach $thisfile ("tfremote-id_dsa", "tfsupport-id_dsa", "id_dsa") {
		if(-f "$userinfo{'homedir'}/.ssh/$thisfile") {
			$id_dsa = "$userinfo{'homedir'}/.ssh/$thisfile";
			open(PIPE, "/usr/bin/md5sum $id_dsa |");
			while(<PIPE>) {
				chomp;
				@array = split(/\s+/);
				if($#array >= 0) {
					$fingerprint .= $array[0];
				}
			}
			close(PIPE);
			last;
		}
	}
	if($id_dsa =~ /tfsupport-id_dsa/) {
		$remoteuser = "tfsupport";
	} else {
		$remoteuser = $localuser;
	}

	if ("$id_dsa" eq "") {
		system("/usr/bin/logger \"$localuser no id_dsa file found.");
		print("Error: Could not find SSH Private Key File.\n");
		return(2);
	}

	unless (-f "$id_dsa") {
		system("/usr/bin/logger \"$localuser $id_dsa File not found.");
		print("Error: SSH Private Key File ($id_dsa) file not found.\n");
		return(3);
	}


	# Log when we start and end our session, as well as who we are.
	system("/usr/bin/logger \"$localuser:$fingerprint:$$ SSH to $remoteuser\@$hostname\"");

	$command = "ssh ";
	$command .= " -t";
	$command .= " -o PubKeyAuthentication=yes";
	$command .= " -o PasswordAuthentication=no";
	$command .= " -o GSSAPIAuthentication=no";
	$command .= " -o StrictHostKeyChecking=no";
	$command .= " -o LogLevel=VERBOSE";
	$command .= " -o DynamicForward=$portnum";
	$command .= " -i $id_dsa";
	$command .= " -l $remoteuser";
	$command .= " $hostname";
	$command .= " \"/d/daisy/bin/tfremote.pl --accept || /bin/bash --login\"";


	print("\n");
	print("\n");
	print("Your SOCKS Port is \"$portnum\"\n");
	print("Configurations should be as follows:\n");
	print("Putty: Connection->SSH->Tunnels->Source port=10000, Destination=localhost:$portnum \n");
	print("FoxyProxy: Manual Proxy Configuration->Host or IP Address=localhost Port=10000\n");
	print("\n");
	print("\n");
	system("$command");
	if($? == 0) {
		system("/usr/bin/logger \"$localuser:$fingerprint:$$ LOGOUT from $remoteuser\@$hostname status SUCCESS\"");
	} else {
		system("/usr/bin/logger \"$localuser:$fingerprint:$$ LOGOUT from $remoteuser\@$hostname status FAIL ($?)\"");
	}

	return(0);
}



#
# When a keypair is created, the public key should force the execution of "tfremote.pl --accept"
# This functionality ensures that we validate password validation, which gives us true, two factor
# authentication. Furthermore, by enforcing password authentication, we also ensure 90 day password
# rotation, all through the "regular" pam stack.
#
sub accept_inbound_connection
{
	my $returnval = 0;
	my $ppid  = undef;
	my %userinfo = ();

	# Setup signal handlers here.
	# $SIG{INT} = die;
	$SIG{STOP} = 'IGNORE';
	%userinfo = get_userinfo(getpwuid($>));
	system("/usr/bin/logger 'Accepted Connection from $userinfo{'username'}.'");

	# This is the 'second factor' of our two-factor authentication.
	$returnval = check_password();
	if($returnval != 0) {
		print("/usr/bin/logger '(2nd factor) Password Validation Failed for $userinfo{'username'}.'\n");
		system("/usr/bin/logger '(2nd factor) Password Validation Failed for $userinfo{'username'}.'");
		$ppid = getppid();
		system("kill -HUP $ppid");
		return(200);
	}


	# What to exec if things go well.
	print("Logged in successfully.\n");
	if (-f "/d/daisy/bin/tfsupport.pl") {
		system("/usr/bin/logger '$userinfo{'username'}: Forcing execution of tfsupport.pl'");
		exec("/d/daisy/bin/tfsupport.pl");
	}

	if (-d "/usr2/bbx/bin") {
		if( ("$userinfo{'username'}" ne "tfsupport") 
		&&  (-f "/usr2/bbx/bin/sshbbx")) {
			system("/usr/bin/logger '$userinfo{'username'}: Forcing execution of tfsupport.pl'");
			exec("/usr2/bbx/bin/sshbbx");
		}
	}

		
	# If all else fails, start the user's shell.... Only for admins.
	system("groups | grep admin > /dev/null 2> /dev/null");
	if($? == 0) {
		system("/usr/bin/logger 'Starting Shell for admin user $userinfo{'username'}.'");
		#exec($userinfo{'shell'});
		exec("/bin/bash");
	}


	# Non admins are SOL at this point.
	$ppid = getppid();
	system("kill -HUP $ppid");
	return(201);
}




# Returns 0 if password was validated,
# Returns non-zero if password failed.
sub check_password
{
	my $command = "";

	system("ssh-keygen -R localhost");
	$command = "ssh ";
	$command .= " -t";
	$command .= " -o StrictHostKeyChecking=no";
	$command .= " -o PubKeyAuthentication=no";
	$command .= " 127.0.0.1 /bin/true";
	system("$command");
	return($?);
}


#
# Allow a user to have remote access to the server.
#
# That means:
# -- verify that the username is not 'tfsupport' or 'daisy', 'rti' or 'root'
# -- verify that the username is in the tfremote group
# -- make public/private key pair for user
# Note that there is preliminary support for "IP Address binding" as well.
# Ticket 9757
#
sub enable_tfremote
{
	my $username = $_[0];
	my @ipaddrs = @{$_[1]};
	my $homedir = "";
        my %userinfo = ();
	my $datestamp = strftime("%Y-%m-%d", localtime());


	if ( (! $username) || ($username eq "") ) {
		print("Error: No username specified to add.\n");
		return(1);
	}
	if ($username eq 'tfsupport') {
		print("Error: This should be done manually for this particular user.\n");
		return(2);
	}

	# PCI-DSS 8.1, 8.2
	# Don't enable remote access for usernsmaes which are clearly "generic."
	# See also http://projecthoneypot.org/dictionary_attacker_usernames.php?vid=tnqc0jae0q8sic6cuiq5cg1900
	if ( ($username eq 'daisy') 
	|| ($username eq 'root')
	|| ($username eq 'tfsupport')
	|| ($username eq 'adm')
	|| ($username eq 'admin')
	|| ($username eq 'administrator')
	|| ($username eq 'odbc')
	|| ($username eq 'manager')
	|| ($username eq 'bookeeper')
	|| ($username eq 'bookkeeper')
	|| ($username eq 'owner')
	|| ($username eq 'clerk')
	|| ($username eq 'pos')
	|| ($username eq 'rti') ) {
		print("Error: Will not enable remote access for default usernames.\n");
		return(3);
	}

	# Is this a valid user?
	print("Enabling tfremote access for $username...\n");
        %userinfo = get_userinfo($username);
        unless (%userinfo) {
		print("User \"$username\" not in /etc/passwd. Skipping.\n");
                return(4);
        }

	# Did our change stick?
	append_usergroup("tfremote", $username);
        %userinfo = get_userinfo($username);
        unless (grep(/^tfremote$/, @{$userinfo{'groups'}}) ) {
                print("Error: \"$username\" is not a tfremote user.\n");
                return(5);
	}


	$homedir = $userinfo{'homedir'};

	print("Creating SSH Keypair for \"$username\"\n");

	# Create ssh subdir if need be.
	unless (-d "$homedir/.ssh") {
		mkdir("$homedir/.ssh");
		system("sudo chown $username:root $homedir/.ssh");
		system("sudo chmod 700 $homedir/.ssh");
	}
	if(! -d "$homedir/.ssh") {
		print("Error: ssh directory does not exist.\n");
		return(6);
	}


	#
	# Create our 'authorized_keys' file, and specify the "login" program
	# to run whenever the user logs in which will mean they have to supply
	# their system account name and passsword to login... thus, this completes
	# the two factor authentication requirement - the key and the password.
	#
	my $ssh_private_key_file = "$homedir/.ssh/tfremote-id_dsa";
	my $ssh_public_key_file  = "$ssh_private_key_file.pub";

	#
	# Make the key pair: type dsa, no passphrase
	#
	system("ssh-keygen -t dsa -f $ssh_private_key_file -C \"$username Remote Access ($datestamp)\" -N \"\"");
	if(! -f "$ssh_private_key_file") {
		print("Error: Could not create ssh private key file \"$ssh_private_key_file\"\n"); 
		return(7);
	}
	if(! -f "$ssh_public_key_file") {
		print("Error: Could not create ssh public key file \"$ssh_public_key_file\"\n"); 
		return(8);
	}


	#
	# Copy the public key and add a command that causes the "login" program to run
	# thus forcing the user to also supply username/password.
	# Note that tfsupport keys are distributed via altiris; and ocntains much more than a single key.
	#
	# Backup our existing 'authorized_keys' file, if that happens to already exist.
	if(-f "$homedir/.ssh/authorized_keys") {
		system("cp $homedir/.ssh/authorized_keys $homedir/.ssh/old-authorized_keys");
		if(! -f "$homedir/.ssh/old-authorized_keys") {
			print("Error: Could not create backup copy of authorized_keys file.\n");
			return(9);
		}
	}

	open(AUTHKEYS, "> $homedir/.ssh/authorized_keys");
	open(PUBKEY, "< $ssh_public_key_file");

	# Retain keys within our existing "authorized_keys" files, or at least, all keys not related to
	# tfremote.
	if(-f "$homedir/.ssh/old-authorized_keys") {

		open(OLDKEYS, "< $homedir/.ssh/old-authorized_keys");
		while(<OLDKEYS>) {
			# Don't replicate older tfremote keys.
			next if(/$username Remote Access/);
			print(AUTHKEYS);
		}
		close(OLDKEYS);
	}

	#
	##############################
	# Start creation of our new public key.
	##############################

	# Restrict this key's access from specified ip address(es).
	# Only if, at setup, we specified an originating IP address.
	if(@ipaddrs) {
		if($#ipaddrs >= 0) {
			print(AUTHKEYS "from=\"");
			foreach(@ipaddrs) {
				print(AUTHKEYS "$_,");
			}
			print(AUTHKEYS "localhost\",");
		}
	}


	if("$username" eq "kiosk") {
		print("Setting up Kiosk SSH Key...\n");
		print(AUTHKEYS  "command=\"/usr/libexec/openssh/sftp-server\",no-port-forwarding,no-X11-forwarding ");
	} else {
		if(-f "/usr2/bbx/bin/tfremote.pl") {
			print(AUTHKEYS  "command=\"/usr2/bbx/bin/tfremote.pl --accept\" ");
		} elsif(-f "/d/daisy/bin/tfremote.pl") {
			print(AUTHKEYS  "command=\"/d/daisy/bin/tfremote.pl --accept\" ");
		} else {
			if("$username" ne "tfsupport") {
				print(AUTHKEYS  "command=\"/bin/echo 'tfremote script not found.'\" ");
			}
		}
	}

	# The rest of the public key.
	while(<PUBKEY>) {
		print(AUTHKEYS);
	}
	close(AUTHKEYS);
	close(PUBKEY);

	system("chown $username:root  $homedir/.ssh/authorized_keys");
	system("chmod 400 $homedir/.ssh/authorized_keys");


	#
	# We have no further use for the public key, so remove it.
	#
	unlink("$ssh_public_key_file");


	#
	# Display authentication credentials.
	# These should be screenscraped and given to the customer.
	#
	system("cat $ssh_private_key_file");



	#
	# Now we can get rid of the private key too.
	#
	unlink("$ssh_private_key_file"); # Don't leave the private key hanging around.


	system("/usr/bin/logger 'User $username successfully added to tfremote.'");

	return(0);
}


#
# Disable the ability for a user to access the system via 'tfremote'.
#
# Ticket 9757
#
sub disable_tfremote
{
	my $username = $_[0];
	my %userinfo = ();

        %userinfo = get_userinfo($username);

	print("Disabling tfremote access for $username...\n");
	system("/usr/bin/logger \"Disabling Remote Access for $username\"");
	unlink("$userinfo{'homedir'}/.ssh/authorized_keys");
	remove_usergroup("tfremote", $username);

	# If kiosk can't get in via remote access, then, there is no need for the kiosk user. Disable him.
	if($username eq "kiosk") {
		if(-f "/usr2/bbx/bin/rtiuser.pl") {
			system("/usr2/bbx/bin/rtiuser.pl --lock kiosk");
		} else {
			system("/usr/bin/passwd -l kiosk");
		}
	}

	return(0);
}
